unit ChessDatabase;

(*

This component was also integrated in TPGN because of its the existing interface. 
The TPGN component uses addGameAsPGN and getGameAsPGN of this unit.
You might use addGame and getGame and compress the game more effectively
(the name of the piece could be used instead of the "from" square,
or maybe, the text fields can be compressed, also).
But in this case you must add the compressing algorithm by yourself.

Data format:

Header:
number (char #3 - #29) + word / string; see the order in DataToPGN
#2 closes the header

Moves:
1 bit  0         : 0 = start of variation, otherwise 1
6 bits 1-6       : From square 0-63
6 bits 7-12       : To square 0-63
1 bit  13        : 0 = move info; 1 = next move
[ 5 Bits 14 - 18 :
              14  0 = next 3 bits : promotion piece
                                    value 0 = N
                                    value 1 = B
                                    value 2 = R
                                    value 3 = Q
              15  0 = next 4 bits : info for Nbd2 or N1d2, for example, and + and #
                                    bit 1 = letter
                                    bit 2 = digit
                                    bit 3 = check
                                    bit 4 = mate
              16  0 = next 8 bits : !, ?, !?, etc., see the values in TNotation
              17  0 = next 8 bits : Chess Informant signs, see the values in TNotation
              18  0 = comment, closed with #1
]
[ x bits from 14        : 0 = end of variation (can be repeated) ]
1 bit 14 [or higher]    : must be 1; closes the variation info
*)

interface

uses Classes, SysUtils, Forms, Controls, UCustomDatabase, Notation,
     SchachProcs, Graphics, Dialogs;

const defaultmainFileExt  = '.ccf';
      defaultidxFileExt   = 'idx';
      defaultstateFileExt = 'tmp';
      ChessDatabaseID = 9876543210;

      cdPromotionKnight = 2;
      cdPromotionBishop = 1;
      cdPromotionRook   = 0;
      cdPromotionQueen  = 3;
      cdVarStart        = $1;
      cdVarEnd          = $2;

      cdLeadingByte     = $1;
      cdCompressedBit   = $80;
      cdFirstPosition   = 255;  // darf nicht als figurwert auf brett enthalten sein

      cdStringLimiter   = #1;
      cdHeaderLimiter   = #2;

      evWElo            = #3;
      evSELO            = #4;
      evPlyCount        = #5;
      evEventDateJahr   = #6;
      evGameSec         = #7;
      evMoveSec         = #8;
      evWeiss           = #9;
      evSchwarz         = #10;
      evquelle          = #11;
      evAnnotator       = #12;
      evort             = #13;
      evergebnis        = #14;
      evEco             = #15;
      evRound           = #16;
      evWhitediffToGMT  = #17;
      evBlackdiffToGMT  = #18;
      evDateTimeFormat  = #19;
      evGameId          = #20;
      evMetaDescription = #21;
      evMetaKeywords    = #22;
      evPosition        = #23;
      evFirstComment    = #24;
      evEventDateMonat  = #25;
      evEventDateTag    = #26;
      evDateJahr        = #27;
      evDateMonat       = #28;
      evDateTag         = #29;
type

  TChessDatabase = class(TCustomDatabase)
    constructor Create(Owner : TComponent);
    destructor destroy; override;
  private
    FmainFileExt, FidxFileExt, FstateFileExt : String;
    FNotation : TNotation;
    function  DataToPGN(Data : String) : String;
    function  PGNToData(PGN : String) : String;
    procedure setmainFileExt(s : String);
    function encodeGame(s : String) : String;
    function decodeGame(s : String) : String;
    procedure doLoadFile(Filename : String; checkDatabaseExtAndSign : Boolean);
  protected
    procedure Paint; override;
    procedure Loaded; override;
  public
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    property idxFileExt : String read FidxFileExt;
    property stateFileExt : String read FstateFileExt;
    (*
    Gets the raw (compressed) game.
    The handle is delivered by addGame, etc.
    Also, deleted games can be got as long as they were not deleted irrecoverably
    by the Compress methos of TCustomDatabase.
    *)
    function  getGame(handle : String; var aGame : String; var recordIsDeleted : Boolean) : Boolean;
    (*
    Gets a raw game by its index in the database from 0..x
    *)
    function  getGameByIndex(idx : Int64; var handle, aGame : String; var recordIsDeleted : Boolean) : Boolean;
    (*
    Gets the game as a pgn string
    *)
    function  getGameAsPGN(handle : String; var PGNGame : String; var recordIsDeleted : Boolean) : Boolean;
    function  getGameAsPGNByIndex(idx : Int64; var handle, PGNGame : String; var recordIsDeleted : Boolean) : Boolean;
    (*
    Replaces a game identified by a handle or the database index.
    You should prefer the handle method if the user could have deleted games in the meantime.
    *)
    function  replaceGame(var handle : String; aGame : String) : Boolean;
    function  replaceGameAsPGN(var handle : String; PGNGame : String) : Boolean;  // the pgn string will not be compressed!
    function  replaceGameByIndex(idx : Int64; aGame : String) : Boolean;
    function  replaceGameAsPGNByIndex(idx : Int64; PGNGame : String) : Boolean;
    (*
    Marks a game as deleted. It is not yet deleted really, only after invoking the "Compress" method.
    *)
    function  deleteGame(handle : String) : Boolean;
    function  deleteGameByIndex(idx : Int64) : Boolean;
    (*
    Adds a game. Your program should remember the handle.
    *)
    function  addGame(var handle : String; aGame : String) : Boolean;
    function  addGameAsPGN(var handle : String; PGNGame : String) : Boolean;
    function  addGameByIndex(var idx : Int64; aGame : String) : Boolean;
    function  addGameAsPGNByIndex(var idx : Int64; PGNGame : String) : Boolean;
    (*
    Returns true if the file is a chess database in the format used by this component
    *)
    function  isChessDatabase(Filename : String) : Boolean;
    (*
    You MUST call one of the next two functions after constructing this component.
    If it was successful, the property "FilesPrepared" of TCustomdatabase is true.
    *)
    procedure LoadFile(Filename : String);
    (*
    Checks the extension and the id of the database before loading
    *)
    procedure CheckAndLoadFile(Filename : String);
  published
    (*
    If Notation is defined, it is used to convert a pgn string to a compressed string
    *)
    property Notation : TNotation read FNotation write FNotation;
    (*
    The extension of the mainfile (default ".ccf")
    In addition, two other files are cretaed with the extensions
    mainFileExt + defaultidxFileExt (".ccfidx") and
    mainFileExt + defaultstateFileExt (".ccftmp")
    Only the latter file will be kept open -
    this component checks it to avoid double access at the same time by various programs
    *)
    property mainFileExt : String read FmainFileExt write setmainFileExt;
  end;

  procedure GetListEx(sl : TStringList; s : string; c : char; maxFields : Integer);

implementation

procedure GetListEx(sl : TStringList; s : string; c : char; maxFields : Integer);
var sh : String;
    i : Integer;
begin
  sh := '';
  For i := 1 to length(s) do begin
    if s[i] = c then begin
      sl.Add(sh);
      if (maxFields > 0) and (sl.Count > maxFields) then
        exit;
      sh := '';
    end
    else
      sh := sh + s[i];
  end;
  sl.Add(sh);
end;

var
  aBitmap: TBitmap;
  Loaded: Boolean;

procedure LoadBitmap;
begin
  if Loaded then exit;
  Loaded := True;
  if aBitmap = nil then aBitmap := TBitmap.Create;
  try
    aBitmap.LoadFromResourceName(HInstance, 'TMINIMAX');
  except
    on E:Exception do ShowMessage(E.Message);
  end;
end;

procedure TChessDatabase.Paint;
begin
  if csDesigning in ComponentState then
    Canvas.StretchDraw(Canvas.ClipRect, aBitmap);
  inherited Paint;
end;

procedure TChessDatabase.Loaded;
begin
  if csDesigning in ComponentState then LoadBitmap
  else SetBounds(0,0,0,0);
  inherited Loaded;
end;

procedure TChessDatabase.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  if csDesigning in ComponentState then
  begin
    AWidth := 25;
    AHeight := 25;
  end;
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
end;

constructor TChessDatabase.Create(Owner : TComponent);
begin
  inherited Create(Owner);
  mainFileExt  := defaultmainFileExt   ;
end;

destructor TChessDatabase.destroy;
begin
  inherited destroy;
end;

procedure TChessDatabase.setmainFileExt(s : String);
var ext : String;
begin
  ext := s;
  if (ext <> '') and (ext[1] <> '.') then
    ext := '.' + ext;
  FmainFileExt  := ext   ;
  FidxFileExt   := ext + defaultidxFileExt    ;
  FstateFileExt := ext + defaultstateFileExt  ;
end;

procedure TChessDatabase.doLoadFile(Filename : String; checkDatabaseExtAndSign : Boolean);
var e1, e2 : String;
begin
  e1 := idxFileExt;
  e2 := stateFileExt;
  if checkDatabaseExtAndSign then begin
    if (FileExists(Filename)) and
    (not isChessDatabase(Filename)) then
      e2 := e1;  // if the extensions are equal, filesprepared is set, but mainFilename will be '' - invalid database
  end;
  maxCheckSec := 30;  // max. 30 sec. database check; can be cancelled in OnDatabaseCheckTrace, also
  prepareFilesEx(Filename, e1, e2, ChessDatabaseID, 0, 0, 0);
end;

procedure TChessDatabase.LoadFile(Filename : String);
begin
  doLoadFile(Filename, false);
end;

procedure TChessDatabase.CheckAndLoadFile(Filename : String);
begin
  doLoadFile(Filename, true);
end;

function TChessDatabase.isChessDatabase(Filename : String) : Boolean;
var s : String;
    f : int64File;
    i : Int64;
begin
  Result:= false;
  if uppercase(ExtractFileExt(Filename)) <> uppercase(mainFileExt) then
    exit;
  s := ChangeFileExt(Filename, idxFileExt);
  if FileExists(s) then begin
    AssignAndResetInt64File(f, s);
    try
      read(f, i);
      if i <> idxID then begin // probably not a chess database, but check the chess id, also
        seek(f, idxUser1);
        read(f, i);
        if i = ChessDatabaseID then
          Result:= true;
      end
      else Result:= true;
      closeFile(f);
    except
      try
        closeFile(f);
      except
      end;
    end;
  end
  else
    Result := true;  // a new index file will be created then
end;

function  TChessDatabase.getGame(handle : String; var aGame : String; var recordIsDeleted : Boolean) : Boolean;
begin
  Result := get(handle, aGame, 0, true, recordIsDeleted);
  if Result then
    aGame := decodeGame(aGame);
end;

function  TChessDatabase.getGameByIndex(idx : Int64; var handle, aGame : String; var recordIsDeleted : Boolean) : Boolean;
begin
  Result := getByIndex(idx, handle, aGame, 0, false, true, -1, recordIsDeleted) = 0;
  if Result then
    aGame := decodeGame(aGame);
end;

function  TChessDatabase.getGameAsPGN(handle : String; var PGNGame : String; var recordIsDeleted : Boolean) : Boolean;
var Data : String;
begin
  PGNGame := '';
  Result := getGame(handle, Data, recordIsDeleted);
  if Result then
    PGNGame := DataToPGN(Data);
end;

function  TChessDatabase.getGameAsPGNByIndex(idx : Int64; var handle, PGNGame : String; var recordIsDeleted : Boolean) : Boolean;
var Data : String;
begin
  PGNGame := '';
  Result := getGameByIndex(idx, handle, Data, recordIsDeleted);
  if Result then
    PGNGame := DataToPGN(Data);
end;

function TChessDatabase.DataToPGN(Data : String) : String;
var s, s1, moves : String;
    stllg : TSchachbrett;
//    sl : TStringlist;
    i, j, idx, step : Integer;
    ev : TEventData;
    b, b1 : Byte;
    bit : Boolean;
    feld1, feld2 : shortint;
    attr1, attr2, UFigur : Integer; //711 shortint;  
    infAttr : byte;
    reserved : longint;
    varinfo : Byte;
    comment : String;
    UFigurAvailable, attr1Available, attr2Available,
    infAttrAvailable, varinfoAvailable, reservedAvailable, commentAvailable : Boolean;
    stepCount : Integer;
    fen : String;
    Zug : TZug;
    znr : word;
    pstllg : PSchachbrett;
    lStack, lStack1 : TList;
    laststllg, rstllg : TSchachbrett;
    idxJustInc : Boolean;
    varEndCount, varStartCount : Integer;
    FirstComment : String;
    infBefore, infBehind : Byte;

const fignames : String = 'RBNQK';

  procedure setStep(i : Integer);
  var sStrich, sVonFeld, sNachFeld, s : String;
      Figur : shortint;
      j : Integer;
  begin
    stepCount := 0;
    step := i;
    case i of
     50 : varStartCount := 0;
      1 : begin
            UFigur := -1;  // stimmt nicht mit normalen figurennamen berein, um 2 Bits (0-3) einzuhalten
            attr1 := 0;
            attr2 := 0;
            infattr := 0;
            varinfo := 0;
            reserved := 0;
            Comment := '';
            Feld1 := 0;
          end;
      2 : Feld2 := 0;
      3 : begin
            UFigurAvailable := false; attr1Available := false;
            attr2Available := false; infAttrAvailable := false;
            varinfoAvailable := false; reservedAvailable := false; commentAvailable := false;
          end;
      4 : UFigur := 0;
      5 : attr1 := 0;
      6 : attr2 := 0;
      7 : infattr := 0;
      8 : Comment := '';
      9 : varEndCount := 0;
      10: begin
            if moves = '' then begin
              moves := moves + IntToStr(znr) + '.';
              if Zug.amZug <> WEISS then
                moves := moves + '..';
            end;
            moves := moves + ' ';
            sStrich := '';

            if varStartCount <> 0 then begin
              New(pstllg);
              pstllg^ := rstllg;
              lStack.Add(pstllg);
              New(pstllg);
              pstllg^ := stllg;
              lStack1.Add(pstllg);
              moves := moves + '(';
              stllg := rstllg;
              varStartCount := 0;
            end
            else rstllg := stllg;

            Figur := abs(Stllg[feld1]);
            if figur = 0 then
              figur := wT;  // rook; should only occur after a castling was ont completely executed

            if (Figur = wK)
            and (
              (
                (feld1 = e1) and
                ((feld2 = g1) or (feld2 = c1))
              ) or
              (
                (feld1 = e8) and
                ((feld2 = g8) or (feld2 = c8))
              )
            ) then begin
               if (feld2 = g1) or (feld2 = g8) then
                 moves := moves + 'O-O'
               else
                 moves := moves + 'O-O-O'
            end
            else begin
              if (figur - 1 > 0) and (figur - 1 <= length(fignames)) then
                moves := moves + fignames[Figur - 1]
              else sStrich := '-';
              sVonFeld := feldnamen[cc(feld1)];
              sNachFeld := feldnamen[cc(feld2)];
              if sStrich = '-' then begin
                if (sVonFeld <> '') and (sNachFeld <> '')
                and (sVonFeld[1] <> sNachFeld[1]) then
                  moves := moves + sVonFeld[1] + 'x' + sNachFeld
                else
                  moves := moves + sVonFeld + '-' + sNachFeld;
              end
              else begin
                s := '';
                if length(sVonFeld) = 2 then begin
                  if attr1 and nLetter <> 0 then
                    moves := moves + sVonFeld[1];
                  if attr1 and nDigit <> 0 then
                    moves := moves + sVonFeld[2];
                  moves := moves + feldnamen[cc(feld2)];
                end;
              end;
            end;

            if (ufigur + 1 > 0) and (ufigur + 1 <= length(fignames)) then begin
              moves := moves + '=' + fignames[uFigur + 1];
              Stllg[feld2] := uFigur + 2;  // nicht +1 wegen verschiebung wegen reduz. von ufigur auf 2 bits
              if Stllg[feld1] < 0 then
                Stllg[feld2] := -Stllg[feld2];
            end
            else
              Stllg[feld2] := Stllg[feld1];
            Stllg[feld1] := LEER;  // the rook position after castling is ignored - if a square should be empty, we assume that a rook was placed there
            s := '';
            if attr1 and nSchach <> 0 then
              s := '+'
            else if attr1 and nMatt <> 0 then
              s := '#';
            moves := moves + s;
            if attr2 <> 0 then
              moves := moves + ' ' + Notation.Attr2toPGNString(attr2);
            if infattr <> 0 then begin
              infBefore := 0;
              infBehind := infattr;    // die umwandlungen sind ntig, weil die nummern fr die beiden Symbole vor den Zgen in die hchstwertigen bits ausgelagert wurden
              if infBehind and 128 <> 0 then begin
                infBefore := infmitderidee;
                infBehind := infBehind or 128 xor 128;
              end;
              if infBehind and 64 <> 0 then begin
                infBefore := infBesserIst;
                infBehind := infBehind or 64 xor 64;
              end;
              case infBehind of
               infgegenspielEx : infBehind := infgegenspiel;
               infzeitnotEx    : infBehind := infzeitnot;
               infneuerungEx   : infBehind := infneuerung;
              end;
              if infBehind <> 0 then
                moves := moves + ' ' + Notation.infAttrtoPGNString(infBehind);
              if infBefore <> 0 then
                moves := moves + ' ' + Notation.infAttrtoPGNString(infBefore);
            end;
            if comment <> '' then
              moves := moves + ' ' + comment;
            if varEndCount <> 0 then begin
              for j := 1 to varEndCount do begin
                moves := moves + ') ';

                if (lStack.Count > 0) and (lStack1.Count > 0) then begin
                  pstllg := lStack[lStack.count - 1];
                  rstllg := pstllg^;
                  lStack.delete(lStack.count - 1);
                  Dispose(pstllg);
                  pstllg := lStack1[lStack1.count - 1];
                  stllg := pstllg^;
                  lStack1.delete(lStack1.count - 1);
                  Dispose(pstllg);
                end;
              end;
              varEndCount := 0;
            end;
            moves := moves + ' ';
            setStep(50);
          end;
    end;
  end;

  function g(s : String) : String;
  begin
    if trim(s) = '' then Result := '?'
    else Result := s;
    Result := '"' + s + '"]' + #13#10;
  end;

  function getNextWord : Word;
  begin
    Result  := 0;
    if i + 2 <= length(s) then
      Result := ((Byte(s[i + 1]) - 1) shl 8) or (Byte(s[i + 2]) - 1);
    inc(i, 2);
  end;

  function getNextString : String;
  begin
    Result  := '';
    inc(i);
    while (i <= length(s)) and (s[i] <> cdStringLimiter) do begin
      Result := Result + s[i];
      inc(i);
    end;
  end;

  function FillZero(s : String) : String;
  begin
    Result := s;
    while length(Result) < 2 do Result := '0' + Result;
  end;

begin
  Result := '';
  if Data = '' then exit;
  znr := 1;
  Zug.amZug := WEISS;

  FirstComment  := '';
  fen := '';
  lStack := TList.Create;
  lStack1 := TList.Create;
  with ev do try
    FillChar(ev, SizeOf(ev), 0);
//    s := DecodeGame(Data);
    s := Data;

    if (s <> '') and (s[1] = '[') then begin  // not compressed, but pgn
      Result := s;
      exit;
    end;  

    i := 1;
    while i <= length(s) do begin
    case s[i] of
      evWElo              : WElo          := getNextWord;
      evSELO              : SELO          := getNextWord;
      evPlyCount          : PlyCount      := getNextWord;
      evEventDateJahr     : EventDateJahr := getNextWord;
      evGameSec           : TimeControlGameSec:= getNextWord;
      evMoveSec           : TimeControlMoveSec:= getNextWord;
      evEventDateMonat    : EventDateMonat:= getNextWord;
      evEventDateTag      : EventDateTag  := getNextWord;
      evDateJahr          : Jahr          := getNextWord;
      evDateMonat         : Monat         := getNextWord;
      evDateTag           : ev.Tag        := getNextWord;
      evRound             : Runde         := getNextWord;
      evWeiss             : ev.Weiss      := getNextString;
      evSchwarz           : Schwarz       := getNextString;
      evquelle            : quelle        := getNextString;  
      evAnnotator         : Annotator     := getNextString;  
      evort               : ort           := getNextString;  
      evergebnis          : ergebnis      := getNextString;
      evEco               : Eco           := getNextString;
      evWhitediffToGMT    : WhitedifferenceToGMTValue:= getNextString;
      evBlackdiffToGMT    : BlackdifferenceToGMTValue:= getNextString;
      evDateTimeFormat    : DateTimeFormatValue:= getNextString;
      evGameId            : GameIdValue        := getNextString;
      evMetaDescription   : MetaDescription:= getNextString;
      evMetaKeywords      : MetaKeywords   := getNextString;
      evPosition          : fen := getNextString;
      evFirstComment      : FirstComment := getNextString;
      cdHeaderLimiter     : break;
      end;
    inc(i);
    end;
    s := copy(s, i + 1, length(s));

    if fen = '' then
      Stllg := Ausgangsstellung
    else
      FenToPos(fen, stllg, zug, znr, Nil);
    rstllg := stllg;

    moves := '';
    setstep(50);
    idx := 0;
    b := 0;
    b1 := 0;
    while true do begin
      idxJustInc := false;
      if b1 = 0 then begin
        idxJustInc := true;
        inc(idx);
        if idx > length(s) then
          break;
        b1 := $80;
        b := Byte(s[idx]);
      end;
      bit := b and b1 <> 0;
      if step = 50 then begin
        if not bit then
          varStartCount := 1;
        setStep(1);
      end
      else if (step = 1) or (step = 2) then begin
        if bit then begin
          if step = 1 then
            Feld1 := Feld1 or ($80 shr (stepCount + 2))
          else
            Feld2 := Feld2 or ($80 shr (stepCount + 2));
        end;
        inc(stepCount);
        if stepCount >= 6 then begin
          if step = 1 then
            setStep(2)
          else begin
            if (Feld1 < 0) or (Feld1 > 63) then exit;
            if (Feld2 < 0) or (Feld2 > 63) then exit;
            feld1 := c(feld1);
            feld2 := c(feld2);

            setStep(3);
          end;
        end;
      end
      else if step = 3 then begin
        if not bit then begin  // zuginfo
          case stepCount of
            1 : UFigurAvailable := true;
            2 : attr1Available := true;
            3 : attr2Available := true;
            4 : infAttrAvailable := true;
            5 : commentAvailable := true;
          end;
        end
        else begin
          if StepCount = 0 then  // if the first bit is 0, no move info follows
            stepCount := 5;
        end;
        inc(stepCount);
        if stepCount >= 6 then begin
          if UFigurAvailable then setStep(4)
          else if attr1Available then setStep(5)
          else if attr2Available then setStep(6)
          else if infAttrAvailable then setStep(7)
          else if commentAvailable then setStep(8)
          else setStep(9);
        end;
      end
      else if step = 4 then begin
        if bit then
          UFigur := UFigur or ($80 shr (stepCount + 6));
        inc(stepCount);
        if stepCount >= 2 then begin
          if attr1Available then setStep(5)
          else if attr2Available then setStep(6)
          else if infAttrAvailable then setStep(7)
          else if commentAvailable then setStep(8)
          else setStep(9);
        end;
      end                  
      else if step = 5 then begin
        if bit then
          attr1 := attr1 or ($80 shr stepCount);
        inc(stepCount);
        if stepCount >= 4 then begin
          if attr2Available then setStep(6)
          else if infAttrAvailable then setStep(7)
          else if commentAvailable then setStep(8)
          else setStep(9);
        end;
      end
      else if step = 6 then begin
        if bit then
          attr2 := attr2 or ($80 shr stepCount);
        inc(stepCount);
        if stepCount >= 8 then begin
          if infAttrAvailable then setStep(7)
          else if commentAvailable then setStep(8)
          else setStep(9);
        end;
      end
      else if step = 7 then begin
        if bit then
          infattr := infattr or ($80 shr stepCount);
        inc(stepCount);
        if stepCount >= 8 then begin
          if commentAvailable then setStep(8)
          else setStep(9);
        end;
      end
      else if step = 8 then begin
      // filled with zero before the comment starts
        if idxJustInc then begin
          while idx <= length(s) do begin
            if s[idx] = cdStringLimiter then
              break;
            if s[idx] = cdHeaderLimiter then begin // for possible extensions
              inc(idx);
              while (idx <= length(s)) and (s[idx] <> cdHeaderLimiter) do
                inc(idx);
            end
            else
              Comment := comment + s[idx];
            inc(idx);
          end;
          setStep(9);
          b1 := 1;
        end;
      end
      else if step = 9 then begin
        if not bit then
          inc(VarEndCount)
        else
          setStep(10);
      end;

      b1 := b1 shr 1;
    end;
    with ev do begin
      Result :=
      '[Event ' + g(quelle) +
      '[Site '  + g(ort) +
      '[Date "';
      if Jahr <> 0 then Result := Result + IntToStr(Jahr)
      else Result := Result + '????';
      Result := Result + '.';
      if Monat <> 0 then Result := Result + FillZero(IntToStr(Monat))
      else Result := Result + '??';
      Result := Result + '.';
      if ev.Tag <> 0 then Result := Result + FillZero(IntToStr(ev.Tag))
      else Result := Result + '??';
      Result := Result + '"]' + #13#10;
      Result := Result + '[Round "';
      if Runde <> 0 then Result := Result + IntToStr(Runde)
      else Result := Result + '-';
      Result := Result + '"]' + #13#10;
      Result := Result +
      '[White ' + g(Weiss) +
      '[Black ' + g(Schwarz) +
      '[Result '+ g(ergebnis);
      if Annotator <> '' then Result := Result + '[Annotator ' + g(Annotator);
      if ECO <> '' then Result := Result + '[ECO ' + g(Eco);
      if WhiteDifferenceToGMTValue <> '' then Result := Result + '[WhiteDifferenceToGMT ' + g(WhitedifferenceToGMTValue);
      if BlackdifferenceToGMTValue <> '' then Result := Result + '[BlackDifferenceToGMT ' + g(BlackdifferenceToGMTValue);
      if DateTimeFormatValue <> '' then Result := Result + '[DateTimeFormat ' + g(DateTimeFormatValue);
      if GameIdValue <> '' then Result := Result + '[GameId ' + g(GameIdValue);
      if fen <> '' then begin
        Result := Result + '[FEN ' + g(fen);
        Result := Result + '[Setup ' + g('1');
      end;


      if WElo <> 0 then Result := Result + '[WhiteElo ' + g(IntToStr(WElo));
      if SElo <> 0 then Result := Result + '[BlackElo ' + g(IntToStr(SElo));
      if PlyCount <> 0 then Result := Result + '[PlyCount ' + g(IntToStr(PlyCount));

      Result := Result + '[EventDate "';
      if EventDateJahr <> 0 then Result := Result + IntToStr(EventDateJahr)
      else Result := Result + '????';
      Result := Result + '.';
      if EventDateMonat <> 0 then Result := Result + FillZero(IntToStr(EventDateMonat))
      else Result := Result + '??';
      Result := Result + '.';
      if EventDateTag <> 0 then Result := Result + FillZero(IntToStr(EventDateTag))
      else Result := Result + '??';
      Result := Result + '"]' + #13#10;
      if TimeControlGameSec <> 0 then begin
        s := IntToStr(TimeControlGameSec);
        if TimeControlMoveSec <> 0 then
          s := s + '+' + IntToStr(TimeControlMoveSec);
        Result := Result + '[TimeControl ' + g(s);
      end;

      Result := Result + #13#10 + FirstComment + ' ' + moves + ' ' + Ergebnis;
    end;
  finally
    lStack.Free;
    lStack1.Free;
  end;
end;

function TChessDatabase.decodeGame(s : String) : String;
begin
  Result := s;
  if Result = '' then exit;
  if Byte(s[1]) and cdCompressedBit <> 0 then
    Result := DecodeData(Result);
  system.Delete(Result, 1, 1);
end;

function TChessDatabase.encodeGame(s : String) : String;
begin
  if Pos(#0, s) <> 0 then
    Result := Char(cdLeadingByte or cdCompressedBit) + EncodeData(s)  // erstes Bit im ersten byte bei codierten datenstzen setzen
  else
    Result := Char(cdLeadingByte) + s;
end;

function TChessDatabase.PGNToData(PGN : String) : String;
var s : String;
begin
  if (FNotation <> Nil) and (FNotation.setGame(PGN) = 0) then begin
    Result := FNotation.getCompressedGame;
  end
  else
    Result := trim(PGN);
end;

// handle is only changed if game could not be replaced, but had to be added
function  TChessDatabase.replaceGame(var handle : String; aGame : String) : Boolean;
begin
  Result := replace(handle, encodeGame(aGame), true);
end;

// if the game could not be replaced here, the game is not added instead
function TChessDatabase.replaceGameByIndex(idx : Int64; aGame : String) : Boolean;
begin
  Result := replaceByIndex(idx, encodeGame(aGame), true);
end;

function  TChessDatabase.replaceGameAsPGN(var handle : String; PGNGame : String) : Boolean;  // the pgn string will not be compressed!
var s : String;
begin
  Result := false;
  s := PGNToData(PGNGame);
  if s = '' then exit;
  Result := replaceGame(handle, s);
end;

function TChessDatabase.replaceGameAsPGNByIndex(idx : Int64; PGNGame : String) : Boolean;
var s : String;
begin
  Result := false;
  s := PGNToData(PGNGame);
  if s = '' then exit;
  Result := replaceGameByIndex(idx, s);
end;

function  TChessDatabase.deleteGame(handle : String) : Boolean;
begin
  Result := delete(handle);
end;

function TChessDatabase.deleteGameByIndex(idx : Int64) : Boolean;
begin
  Result := deleteByIndex(idx);
end;

function  TChessDatabase.addGame(var handle : String; aGame : String) : Boolean;
begin
  Result := add(handle, encodeGame(aGame), true);
end;

function  TChessDatabase.addGameByIndex(var idx : Int64; aGame : String) : Boolean;
var dummyhandle : String;
begin
  idx := -1;
  Result := addGame(dummyhandle, aGame);
  if Result then
    idx := CountWithDeletedRecords - 1;
end;

function  TChessDatabase.addGameAsPGN(var handle : String; PGNGame : String) : Boolean;
var s : String;
begin
  Result := false;
  s := PGNToData(PGNGame);
  if s = '' then exit;
  Result := addGame(handle, s);
end;

function  TChessDatabase.addGameAsPGNByIndex(var idx : Int64; PGNGame : String) : Boolean;
var s : String;
   test1 , test2, testhandle : String;
   testdel : Boolean;
begin
  Result := false;
  s := PGNToData(PGNGame);
  if s = '' then exit;
  Result := addGameByIndex(idx, s);

//  check ...
(*  FNotation.SetGame(PGNGame);
  test1 := FNotation.getGame;
  test2 := '';
  if getGameaspgnbyIndex(idx, testhandle, test2, testdel) then begin
    FNotation.checkMoves := true;
    if FNotation.SetGame(test2) <> 0 then
      Result := true;
    FNotation.checkMoves := false;
    test2 := FNotation.getGame;
  end;
  if test1 <> test2 then
   Result := true;*)
end;

initialization
  aBitmap := nil;
  Loaded := False;
finalization
  aBitmap.Free;

end.

